#include <math.h>
#include <stdio.h>

/* This is how I chose to view the complex numbers.  You can */
/* also use two arrays of doubles, one for the real and one  */
/* for the imaginary part.  There are probably also other    */
/* ways of doing this. */

typedef struct Complex{
  double re;
  double im;
} complex;

/******************/
/*** init_array ***/
/******************/

/*
 * IN : n - dimension of A
 *      A - an array of n^3 complex elements
 *      gen_re - a pointer to a function which generates the real part 
 *               of the input data
 *      gen_im - a pointer to a function which generates the imaginary 
 *               part of the input data
 * OUT:
 * MOD: A - entries filled with values determined by gen_re and gen_im
 * REQ: n positive integer
 */

void init_array(int n, complex *A, double (*gen_re)(int, int, int, int),
		double (*gen_im)(int, int, int, int))
{
  int i, j, k;

  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) {
      for (k=0; k<n; k++) {
	A[i*n*n + j*n + k].re = (*gen_re)(n,i,j,k);
	A[i*n*n + j*n + k].im = (*gen_im)(n,i,j,k);
      }
    }
  }
}

/********************/
/*** output_array ***/
/********************/

/*
 * IN : n - dimension of A
 *      outputfile - name of the file to write output to
 *      A - an array of n^3 complex elements
 * OUT:
 * MOD: a file with name "outputfile" will be opened.  It's first
 *        line will have the value of n.  Subsequent lines will 
 *        contain entries of A.
 *      stderr
 * REQ: n positive integer
 */

void output_array(int n, char *outputfile, complex *A)
{
  int i, j, k;
  FILE *fp1;

  fp1 = fopen(outputfile, "w");

  if (fp1 == NULL) {
    printf("could not open %s for output\n", outputfile);
    exit(-1);
  }
  fprintf(fp1, "%d\n", n);

  for (i=0; i<n; i++)
    for (j=0; j<n; j++)
      for (k=0; k<n; k++) {
/*	fprintf(fp1, "%g %g\n", A[i*n*n+j*n+k].re, A[i*n*n+j*n+k].im); */
	fprintf(fp1, "%lf %lf\n", A[i*n*n+j*n+k].re, A[i*n*n+j*n+k].im);
      }

  fclose(fp1);
}

/*****************/
/*** compute_W ***/
/*****************/

/*
 * IN : n - dimension of W
 *      W - array of n^2 complex numbers
 * OUT:
 * MOD: W - will contain roots of unity so that W[i*n+j] = e^(2*pi*i*j/n)
 *      stderr
 * REQ: n positive integer
 */

void compute_W(int n, complex *W)
{
  int i, j;
  double t1, t2;
  complex *RU;

  RU = (complex *) malloc(n * sizeof(complex));
  if (RU == NULL) {
    fprintf(stderr,"Couldn't allocate work space RU in compute_W\n");
    exit(-1);
  }

  for (i=0; i<n; i++) {
    t1 = cos(((double)i*2.0*M_PI)/((double)n));
    t2 = sin(((double)i*2.0*M_PI)/((double)n));
    RU[i].re = t1;
    RU[i].im = t2;
  }

  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) {
      W[i*n+j].re = RU[(i*j)%n].re;
      W[i*n+j].im = RU[(i*j)%n].im;
    }
  }

  free(RU);
}

/*******************/
/*** naive_fft1d ***/
/*******************/

/*
 * IN : n - dimension of W, length of vector to take the dft of.
 *      in - pointer to the first element of the vector we're going
 *           to take the dft of.
 *      out - complex array of size n
 *      W - array of n^2 complex numbers as defined by compute_W
 *      stride - distance between elements of the vector in.
 * OUT:
 * MOD: out - contains the dft of the vector in.
 * REQ: n positive integer
 *      in - the full array from which the input in comes from must
 *           have enough elements so that in+(stride*(n-1)) remains
 *           a valid element.
 */

void naive_fft1d(int n, complex *in, complex *out, complex *W, int stride)
{
  int i, j;
  complex temp;

  for (i=0; i<n; i++) {
    temp.re = 0.0;
    temp.im = 0.0;
    for (j=0; j<n; j++) {
      temp.re += (W[i*n+j].re * in[j*stride].re) - 
	(W[i*n+j].im * in[j*stride].im);
      temp.im += (W[i*n+j].re * in[j*stride].im) + 
	(W[i*n+j].im * in[j*stride].re);
    }
    out[i].re = temp.re;
    out[i].im = temp.im;
  }
}

/*******************/
/*** naive_fft2d ***/
/*******************/

/*
 * IN : n - dimension of W, dimension of the matrix to take the 2d dft of.
 *      in - pointer to the first element of the 2d matrix we're going
 *           to take the 2d dft of.
 *      W - array of n^2 complex numbers as defined by compute_W
 *      stride1 - distance between elements of the row vectors of the
 *                matrix we're taking the 2d dft of.
 *      stride2 - distance between elements of the column vectors of 
 *                the matrix we're taking the 2d dft of.
 * OUT:
 * MOD: in - overwritten by the 2d dft of the original in.
 *      stderr
 * REQ: n positive integer
 *      in - the full array from which the input in comes from must
 *           have enough elements so that in+(stride1*(n-1))+(stride2*(n-1)) 
 *           remains a valid element.
 *
 * NOTES: takes the 2d dft by taking 1d dfts in one direction and then
 *        1d dfts in the other direction.  assumes naive_fft1d works.
 */

void naive_fft2d(int n, complex *in, complex *W, int stride1, int stride2)
{
  int i, j;
  complex *current_out;

  /* allocating work space used in the 1d ft */
  current_out = (complex *) malloc(n * sizeof(complex));
  if (current_out == NULL) {
    fprintf(stderr,
	    "couldn't allocate work space current_out in naive_fft2d\n");
    exit(-1);
  }

  /* first do 1D fts in 1 direction */
  for (i=0; i<n; i++) {
    naive_fft1d(n, &in[stride2*i], current_out, W, stride1);
    for (j=0; j<n; j++) {
      in[stride2*i+stride1*j].re = current_out[j].re;
      in[stride2*i+stride1*j].im = current_out[j].im;
    }
  }

  /* now do the 1D fts in the other direction */
  for (i=0; i<n; i++) {
    naive_fft1d(n, &in[stride1*i], current_out, W, stride2);
    for (j=0; j<n; j++) {
      in[stride1*i+stride2*j].re = current_out[j].re;
      in[stride1*i+stride2*j].im = current_out[j].im;
    }
  }
}


/*******************/
/*** naive_fft3d ***/
/*******************/

/* 
 * IN : n - dimension of the matrix to take the 3d dft of
 *      gen_re - pointer to a function which takes n, i, j, and k 
 *               and returns the real part of the initial value of 
 *               A[i*n^2 + j*n + k] 
 *      gen_im - pointer to a function which takes n, i, j, and k 
 *               and returns the imaginary part of the initial value 
 *               of A[i*n^2 + j*n + k] 
 *      outfile - name of an output file
 * OUT:
 * MOD: a file with the name "outfile" will be opened and written
 *        into.  the first entry will be n, followed by the elements 
 *        of the 3d dft of an n^3 matrix whose original elements are 
 *        determined by gen_re and gen_im.  The output will be in
 *        normal order.
 *      stderr
 * REQ: n positive integer
 */

void naive_fft3d(int n, 
		 double (*gen_re)(int, int, int, int), 
		 double (*gen_im)(int, int, int, int), 
		 char *outfile)
{
  complex *A, *W;
  int i, j, k;
  complex *current_out;

  A = (complex *) malloc(n * n * n * sizeof(complex));
  if (A == NULL) {
    fprintf(stderr,"Couldn't allocate memory for A in naive_fft3d\n");
    exit(-1);
  }
  W = (complex *) malloc(n * n * sizeof(complex));
  if (W == NULL) {
    fprintf(stderr,"Couldn't allocate memory for W in naive_fft3d\n");
    exit(-1);
  }

  /* allocate work space */
  current_out = (complex *) malloc(n * sizeof(complex));
  if (current_out == NULL) {
    fprintf(stderr,
	    "Couldn't allocate workspace for current_out in naive_fft3d\n");
    exit(-1);
  }

  init_array(n, A, gen_re, gen_im);
  compute_W(n, W);

  /* Then a 2D fft so that we do 1D fts along both the x and y axes */
  for (i=0; i<n; i++) {
    naive_fft2d(n, &A[i*n*n], W, 1, n);
  }

  /* first do 1D fts in 1 direction, along the z axis */
  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) {
      naive_fft1d(n, &A[i*n+j], current_out, W, n*n);
      for (k=0; k<n; k++) {
	A[n*n*k+n*i+j].re = current_out[k].re;
	A[n*n*k+n*i+j].im = current_out[k].im;
      }
    }
  }

  output_array(n, outfile, A); 
  free(current_out);
  free(W);
  free(A);
}


